﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text.RegularExpressions;

namespace ConsoleTestApp
{
    public interface IProtocol
    {
        byte[] Encode();

        IProtocol Decode(byte[] sourceBytes);

        List<DataDefine> DataDefines { get;}
    }

    public class DataDefine
    {
        public string Name { get; set; }

        public byte[] Data { get; set; }

        public int Length { get; set; }
    }

    public abstract class BaseProtocol : IProtocol
    {
        protected const byte FrameHead = 0x68;
        
        public byte DeviceType { get; protected set; }

        public byte[] Address { get; protected set; }

        public byte ControlCode { get; protected set; }

        public int DataLength { get; protected set; }

        public byte[] DataArea { get; private set; }

        public List<DataDefine> DataDefines { get;}
        
        public byte AccumulateSum { get; protected set; }

        protected const byte FrameEnd = 0x16;
        
        public BaseProtocol()
        {
            DataDefines = new List<DataDefine>();
        }

        public DataDefine this[string key]
        {
            get
            {
                return DataDefines.FirstOrDefault(x => x.Name == key);
            }
        }

        public abstract byte[] Encode();

        public abstract IProtocol Decode(byte[] sourceBytes);
        
        protected virtual void BuildDataArea()
        {
            // 构建数据域。
            using (var dataMemory = new MemoryStream())
            {
                foreach (var data in DataDefines)
                {
                    if(data==null) continue;
                    dataMemory.Write(data.Data);
                }

                DataArea = dataMemory.ToArray();
                DataLength = DataArea.Length;
            }
        }
    }

    public class CJT188 : BaseProtocol
    {
        public override byte[] Encode()
        {
            // 校验协议数据。
            if(Address.Length != 7) throw new ArgumentException($"水表地址 {BitConverter.ToString(Address)} 的长度不正确，长度不等于 7 个字节。");

            BuildDataArea();
            
            using (var mem = new MemoryStream())
            {
                mem.WriteByte(FrameHead);
                mem.WriteByte(DeviceType);
                mem.Write(Address);
                mem.WriteByte(ControlCode);
                mem.WriteByte((byte)DataLength);
                mem.Write(DataArea);
                AccumulateSum = ByteUtils.CalculateAccumulateSum(mem.ToArray());
                mem.WriteByte(AccumulateSum);
                mem.WriteByte(FrameEnd);

                return mem.ToArray();
            }
        }

        public override IProtocol Decode(byte[] sourceBytes)
        {
            using (var mem = new MemoryStream(sourceBytes))
            {
                using (var reader = new BinaryReader(mem))
                {
                    reader.ReadByte();
                    
                    DeviceType = reader.ReadByte();
                    Address = reader.ReadBytes(7);
                    ControlCode = reader.ReadByte();
                    DataLength = reader.ReadByte();
                    foreach (var dataDefine in DataDefines)
                    {
                        dataDefine.Data = reader.ReadBytes(dataDefine.Length);
                    }

                    AccumulateSum = reader.ReadByte();
                }
            }

            return this;
        }
    }

    public class CJT188_Read_Request : CJT188
    {
        public CJT188_Read_Request(string address,byte type)
        {
            Address = ByteUtils.HexStringToBytes(address).Reverse().ToArray();
            ControlCode = 0x1;
            DeviceType = type;
            
            DataDefines.Add(new DataDefine{Name = "Default",Length = 2});
            DataDefines.Add(new DataDefine{Name = "Seq",Length = 1});
        }
    }

    public class CJT188_Read_Response : CJT188
    {
        public CJT188_Read_Response()
        {
            DataDefines.Add(new DataDefine{Name = "Default",Length = 2});
            DataDefines.Add(new DataDefine{Name = "Seq",Length = 1});
            DataDefines.Add(new DataDefine{Name = "AllData",Length = 4});
            DataDefines.Add(new DataDefine{Name = "AllDataUnit",Length = 1});
            DataDefines.Add(new DataDefine{Name = "MonthData",Length = 4});
            DataDefines.Add(new DataDefine{Name = "MonthDataUnit",Length = 1});
            DataDefines.Add(new DataDefine{Name = "DateTime",Length = 7});
            DataDefines.Add(new DataDefine{Name = "Status1",Length = 1});
            DataDefines.Add(new DataDefine{Name = "Status2",Length = 1});
        }
    }
    
    class Program
    {
        static void Main(string[] args)
        {
            // 发送水表读表数据。
            var sendProtocol = new CJT188_Read_Request("00000805000001",0x10);
            sendProtocol["Default"].Data = new byte[] {0x1F, 0x90};
            sendProtocol["Seq"].Data = new byte[] {0x00};

            Console.WriteLine(BitConverter.ToString(sendProtocol.Encode()));
            
            // 解析水表响应数据。
            var receiveProtocol = new CJT188_Read_Response().Decode(ByteUtils.HexStringToBytes("68 10 78 06 12 18 20 00 00 81 16 90 1F 00 00 01 00 00 2C 00 01 00 00 2C 00 00 00 00 00 00 00 01 FF E0 16"));
            
            Console.ReadLine();
        }
    }

    public static class ByteUtils
    {
        /// <summary>
        /// 计算一组二进制数据的累加和。
        /// </summary>
        /// <param name="waitCalcBytes">等待计算的二进制数据。</param>
        public static byte CalculateAccumulateSum(byte[] waitCalcBytes)
        {
            int ck = 0;
            foreach (var @byte in waitCalcBytes) ck = (ck + @byte);
            // 对 256 取余，获得 1 个字节的数据。
            return (byte)(ck % 0x100);
        }
        
        /// <summary>
        /// 判断输入的字符串是否是有效的十六进制数据。
        /// </summary>
        /// <param name="hexStr">等待判断的十六进制数据。</param>
        /// <returns>符合规范则返回 True，不符合则返回 False。</returns>
        public static bool IsIllegalHexadecimal(string hexStr)
        {
            var validStr = hexStr.Replace("-", string.Empty).Replace(" ", string.Empty);
            if (validStr.Length % 2 != 0) return false;
            if (string.IsNullOrEmpty(hexStr) || string.IsNullOrWhiteSpace(hexStr)) return false;

            return new Regex(@"[A-Fa-f0-9]+$").IsMatch(hexStr);
        }
        
        /// <summary>
        /// 将 16 进制的字符串转换为字节数组。
        /// </summary>
        /// <param name="hexStr">等待转换的 16 进制字符串。</param>
        /// <returns>转换成功的字节数组。</returns>
        public static byte[] HexStringToBytes(string hexStr)
        {
            // 处理干扰，例如空格和 '-' 符号。
            var str = hexStr.Replace("-", string.Empty).Replace(" ", string.Empty);

            return Enumerable.Range(0, str.Length)
                .Where(x => x % 2 == 0)
                .Select(x => Convert.ToByte(str.Substring(x, 2), 16))
                .ToArray();
        }
    }
}